# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import pkg_resources
from hamcrest import equal_to


def open_fixture(filename):
    return open(pkg_resources.resource_filename('netman.api', 'doc_config/api_samples/' + filename), "r")


def matches_fixture(filename):
    return equal_to(json.load(open_fixture(filename)))
